@@ -6,7 +6,7 @@ use EV ();
 use Carp ();
 use Socket ();
 
-our $VERSION = '1.506_05';
+our $VERSION = '1.506_06';
 
 require Feersum::Connection;
 require Feersum::Connection::Handle;
@@ -18,22 +18,39 @@ XSLoader::load('Feersum', $VERSION);
 $VERSION = eval $VERSION; ## no critic (StringyEval, ConstantVersion)
 
 our $INSTANCE;
+my %_SOCKETS; # inside-out storage for socket refs (keyed by Scalar::Util::refaddr)
 
+use Scalar::Util ();
 use Exporter 'import';
 our @EXPORT_OK = qw(HEADER_NORM_UPCASE HEADER_NORM_LOCASE HEADER_NORM_UPCASE_DASH HEADER_NORM_LOCASE_DASH);
 
 sub new {
     unless ($INSTANCE) {
-        $INSTANCE = bless {}, __PACKAGE__;
+        $INSTANCE = __PACKAGE__->_xs_default_server();
         $SIG{PIPE} = 'IGNORE';
     }
     return $INSTANCE;
 }
 *endjinn = *new;
 
+sub new_instance {
+    my $class = shift;
+    $SIG{PIPE} = 'IGNORE';
+    return $class->_xs_new_server();
+}
+
+sub DESTROY {
+    my $self = shift;
+    my $addr = Scalar::Util::refaddr($self);
+    delete $_SOCKETS{$addr};
+    # XS DESTROY is renamed to _xs_destroy and called here
+    $self->_xs_destroy();
+}
+
 sub use_socket {
     my ($self, $sock) = @_;
-    $self->{socket} = $sock;
+    my $addr = Scalar::Util::refaddr($self);
+    push @{$_SOCKETS{$addr}}, $sock; # keep ref to prevent GC
     my $fd = fileno $sock;
     Carp::croak "Invalid socket: fileno returned undef" unless defined $fd;
     $self->accept_on_fd($fd);
@@ -69,6 +86,10 @@ sub use_socket {
 # callbacks.
 sub DIED { Carp::confess "DIED: $@"; }
 
+# TLS convenience: set_tls on the just-added listener
+# Usage: $server->use_socket($sock); $server->set_tls(cert_file => $path, key_file => $path);
+# The XS method set_tls is already available directly.
+
 1;
 __END__
 
@@ -377,7 +398,7 @@ writer is dropped.
 
 =head1 METHODS
 
-These are methods on the global Feersum singleton.
+These are methods on the Feersum server object.
 
 =over 4
 
@@ -387,14 +408,24 @@ These are methods on the global Feersum singleton.
 
 Returns the C<Feersum> singleton. Takes no parameters.
 
+=item C<< new_instance() >>
+
+Creates a new independent Feersum server instance. Unlike C<new()>, each
+call returns a separate server object with its own listeners, configuration,
+and request handler. Use this when you need multiple independent servers in
+the same process.
+
+    my $http  = Feersum->new_instance();
+    my $https = Feersum->new_instance();
+
 =item C<< use_socket($sock) >>
 
 Use the file-descriptor attached to a listen-socket to accept connections.
 
-TLS sockets are B<NOT> supported nor are they detected. Feersum needs to use
-the socket at a low level and will ignore any encryption that has been
-established (data is always sent in the clear).  The intented use of Feersum
-is over localhost-only sockets.
+B<Note:> Pre-encrypted sockets (e.g. L<IO::Socket::SSL>) are not supported.
+Feersum operates on the raw file descriptor and will ignore any userspace
+encryption layer.  To enable TLS, use C<set_tls()> after adding the socket;
+Feersum handles encryption internally via picotls.
 
 A reference to C<$sock> is kept as C<< Feersum->endjinn->{socket} >>.
 
@@ -523,6 +554,19 @@ Override Feersum's notion of what SERVER_HOST and SERVER_PORT should be.
 
 Override Feersum's default keepalive behavior.
 
+=item C<< set_reverse_proxy($bool) >>
+
+Enable or disable reverse proxy mode. When enabled, Feersum trusts
+C<X-Forwarded-For> and C<X-Forwarded-Proto> headers from upstream proxies to
+determine the client's real IP address and request scheme.
+
+The L<Feersum::Connection> methods C<remote_addr()>, C<remote_port()>, and
+C<env()> will automatically use the forwarded values when this mode is active.
+
+=item C<< get_reverse_proxy() >>
+
+Returns whether reverse proxy mode is currently enabled (1 or 0).
+
 =item C<< max_connection_reqs() >>
 
 =item C<< max_connection_reqs($count) >>
@@ -572,6 +616,30 @@ Only effective on Linux systems; has no effect on other platforms.
 
 Returns true if EPOLLEXCLUSIVE mode is enabled, false otherwise.
 
+=item C<< set_proxy_protocol($bool) >>
+
+Enable or disable PROXY protocol support. When enabled, Feersum expects all
+new connections to begin with a PROXY protocol header (v1 text or v2 binary
+format, auto-detected) before any HTTP data.
+
+The PROXY protocol is used by load balancers like HAProxy, AWS ELB/NLB, and
+nginx to pass the original client IP address to backend servers. When a valid
+PROXY header is received, REMOTE_ADDR and REMOTE_PORT are updated to reflect
+the client's real address.
+
+Special cases:
+- PROXY v1 UNKNOWN: Keeps original address (used for health checks)
+- PROXY v2 LOCAL: Keeps original address (used for health checks)
+
+Connections without a valid PROXY header will be rejected with HTTP 400.
+
+B<Only enable this when ALL connections come from a proxy that sends PROXY
+headers.>
+
+=item C<< get_proxy_protocol() >>
+
+Returns true if PROXY protocol support is enabled, false otherwise.
+
 =item C<< max_accept_per_loop() >>
 
 =item C<< max_accept_per_loop($count) >>
@@ -609,6 +677,43 @@ Setting this to 0 disables the limit. In production, consider also running
 Feersum behind a reverse proxy (nginx, HAProxy) which can provide additional
 connection limiting and rate limiting.
 
+=item C<< set_tls(cert_file => $path, key_file => $path, [listener => $idx]) >>
+
+Enable TLS 1.3 on a listener. Requires Feersum to be compiled with TLS
+support (picotls submodule + OpenSSL; see L<Alien::OpenSSL>).
+
+The cert_file should be a PEM-encoded certificate chain, and key_file the
+corresponding PEM-encoded private key.
+
+The optional C<listener> parameter specifies which listener to configure
+(0-based index, in order of C<use_socket()>/C<accept_on_fd()> calls).
+Defaults to the last-added listener.
+
+Call this after C<use_socket()> or C<accept_on_fd()> to apply TLS to that
+listener. Different listeners can have different TLS configurations, or some
+can be plain HTTP while others use TLS.
+
+    my $ngn = Feersum->endjinn;
+    $ngn->use_socket($tls_socket);
+    $ngn->set_tls(cert_file => 'server.crt', key_file => 'server.key');
+
+When TLS is enabled and L<Alien::nghttp2> was available at build time,
+HTTP/2 can be enabled by passing C<< h2 => 1 >> (HTTP/2 is off by
+default):
+
+    $ngn->set_tls(cert_file => 'server.crt', key_file => 'server.key',
+                  h2 => 1);
+
+L<Feersum::Runner> also accepts C<< h2 => 1 >> as a top-level option.
+
+=item C<< has_tls() >>
+
+Returns true if Feersum was compiled with TLS support (picotls).
+
+=item C<< has_h2() >>
+
+Returns true if Feersum was compiled with HTTP/2 support (nghttp2).
+
 =back
 
 =cut
@@ -710,6 +815,68 @@ C<psgix.body.scalar_refs> feature.
 
 =back
 
+=head2 HTTP/2 Support
+
+When Feersum is built with TLS (picotls + L<Alien::OpenSSL>) and HTTP/2
+(L<Alien::nghttp2>) support, HTTP/2 can be negotiated via ALPN on TLS
+connections.  HTTP/2 is B<disabled by default> and must be explicitly
+enabled by passing C<< h2 => 1 >> to C<set_tls()> or to L<Feersum::Runner>.
+
+=over 4
+
+=item *
+
+B<TLS-only> -- cleartext HTTP/2 (h2c) is not supported.  HTTP/2 is
+negotiated exclusively through the C<h2> ALPN token during the TLS
+handshake.
+
+=item *
+
+B<Request methods> -- all standard methods (GET, POST, PUT, DELETE, etc.)
+are supported.  Request bodies are fully buffered before the handler is
+called, same as HTTP/1.x.
+
+=item *
+
+B<Streaming responses> -- the C<psgi.streaming> / C<start_streaming()>
+interface works over HTTP/2, with each C<write()> producing DATA frames.
+
+=item *
+
+B<Multiple concurrent streams> -- the server processes many streams in
+parallel on a single connection, up to C<FEER_H2_MAX_CONCURRENT_STREAMS>
+(default 100).
+
+=item *
+
+B<Not supported> -- server push, trailers, streaming (incremental) request
+bodies, and C<sendfile>.  For HTTP/2 responses, use C<write()> instead of
+C<sendfile>.
+
+=item *
+
+B<PSGI environment> -- C<psgi.url_scheme> is C<https> for HTTP/2 streams.
+C<SERVER_PROTOCOL> is C<HTTP/2>.
+
+=item *
+
+B<Extended CONNECT / WebSocket tunnels (RFC 8441)> -- Feersum advertises
+C<SETTINGS_ENABLE_CONNECT_PROTOCOL=1> so HTTP/2 clients can open WebSocket
+tunnels via Extended CONNECT.  The PSGI handler sees:
+
+    REQUEST_METHOD       => 'CONNECT'
+    HTTP_UPGRADE         => 'websocket'       # synthesised from :protocol
+    psgix.h2.protocol    => 'websocket'       # raw :protocol value
+    psgix.h2.extended_connect => 1
+
+The handler accepts or rejects the tunnel by calling C<start_streaming>
+with a C<200> status (accept) or a C<4xx> status (reject).  Once accepted,
+C<< psgix.io >> (or C<< $req->io() >>) returns a bidirectional handle backed
+by a Unix socketpair; Feersum bridges bytes between that handle and the
+HTTP/2 DATA frames in both directions.
+
+=back
+
 =head1 BUGS
 
 Please report bugs using http://github.com/stash/Feersum/issues/
