@@ -29,27 +29,26 @@ use constant DEATH_TIMER => 5.0; # seconds
 use constant DEATH_TIMER_INCR => 2.0; # seconds
 use constant DEFAULT_HOST => 'localhost';
 use constant DEFAULT_PORT => 5000;
-use constant MAX_PRE_FORK => 1000; # warn if pre_fork exceeds this
+use constant MAX_PRE_FORK => 1000; # croak if pre_fork exceeds this
 
 our $INSTANCE;
 sub new { ## no critic (RequireArgUnpacking)
     my $c = shift;
-    croak "Only one Feersum::Runner instance can be active at a time"
-        if $INSTANCE && $INSTANCE->{running};
-    # Clean up old instance state before creating new one
     if ($INSTANCE) {
-        $INSTANCE->DESTROY();
+        croak "Only one Feersum::Runner instance can be active at a time"
+            if $INSTANCE->{running};
+        # Clean up old instance state before creating new one
+        $INSTANCE->_cleanup();
         undef $INSTANCE;
     }
     $INSTANCE = bless {quiet=>1, @_, running=>0}, $c;
     return $INSTANCE;
 }
 
-sub DESTROY {
-    local $@;
+sub _cleanup {
     my $self = shift;
-    return if $self->{_destroyed};  # Guard against double-DESTROY
-    $self->{_destroyed} = 1;
+    return if $self->{_cleaned_up};
+    $self->{_cleaned_up} = 1;
     if (my $f = $self->{endjinn}) {
         $f->request_handler(sub{});
         $f->unlisten();
@@ -59,6 +58,11 @@ sub DESTROY {
     return;
 }
 
+sub DESTROY {
+    local $@;
+    $_[0]->_cleanup();
+}
+
 sub _create_socket {
     my ($self, $listen, $use_reuseport) = @_;
 
@@ -78,7 +82,7 @@ sub _create_socket {
         my $err = $@;
         umask($saved);  # Restore umask even if socket creation failed
         die $err if $err;
-        croak "couldn't bind to socket: $!" unless $sock;
+        croak "couldn't bind to socket" unless $sock;
         $sock->blocking(0) || do { close($sock); croak "couldn't unblock socket: $!"; };
     }
     else {
@@ -94,13 +98,14 @@ sub _create_socket {
                 # IPv6 without port: [::1]
                 ($host, $port, $is_ipv6) = ($1, 0, 1);
             } elsif ($listen =~ /:.*:/) {
-                # Bare IPv6 (multiple colons): could be ::1 or 2001:db8::1
-                # Reject ambiguous cases - require bracket notation for IPv6 with ports
-                # e.g., ::1:8080 is ambiguous (address ::1 port 8080? or address ::1:8080?)
-                if ($listen =~ /:(\d{1,5})$/ && $1 <= 65535) {
-                    # Looks like it ends with a port number - ambiguous!
-                    croak "ambiguous IPv6 address '$listen': use bracket notation [host]:port " .
-                          "(e.g., [::1]:$1 or [2001:db8::1]:$1)";
+                # Bare IPv6 - reject ambiguous cases that look like host:port
+                if ($listen =~ /:(\d{1,5})$/) {
+                    my $maybe_port = $1;
+                    # 5 digits = definitely a port; >=1024 = likely a port
+                    if ($maybe_port <= 65535 && (length($maybe_port) == 5 || $maybe_port >= 1024)) {
+                        croak "ambiguous IPv6 address '$listen': use bracket notation [host]:port " .
+                              "(e.g., [::1]:$maybe_port or [2001:db8::1]:$maybe_port)";
+                    }
                 }
                 ($host, $port, $is_ipv6) = ($listen, 0, 1);
             } else {
@@ -184,15 +189,13 @@ sub _prepare {
         if ref $self->{listen} ne 'ARRAY';
     croak "listen array cannot be empty"
         if @{$self->{listen}} == 0;
-    croak "Feersum doesn't support multiple 'listen' directives yet"
-        if @{$self->{listen}} > 1;
-    my $listen = shift @{$self->{listen}};
-    $self->{_listen_addr} = $listen;  # Store for children when using reuseport
+    $self->{_listen_addrs} = [ @{$self->{listen}} ];
 
     if (my $opts = $self->{options}) {
         $self->{$_} = delete $opts->{$_} for grep defined($opts->{$_}),
             qw/pre_fork keepalive read_timeout header_timeout max_connection_reqs reuseport epoll_exclusive
-               read_priority write_priority accept_priority max_accept_per_loop max_connections/;
+               read_priority write_priority accept_priority max_accept_per_loop max_connections
+               reverse_proxy proxy_protocol h2 tls tls_cert_file tls_key_file/;
         # Warn about unknown options (likely typos)
         for my $unknown (keys %$opts) {
             carp "Unknown option '$unknown' ignored";
@@ -214,9 +217,6 @@ sub _prepare {
     my $use_reuseport = $self->{reuseport} && $self->{pre_fork} && defined SO_REUSEPORT;
     $self->{_use_reuseport} = $use_reuseport;
 
-    my $sock = $self->_create_socket($listen, $use_reuseport);
-    $self->{sock} = $sock;
-
     my $f = Feersum->endjinn;
 
     # EPOLLEXCLUSIVE must be set BEFORE use_socket() so the separate accept epoll
@@ -225,9 +225,19 @@ sub _prepare {
         $f->set_epoll_exclusive(1);
     }
 
-    $f->use_socket($sock);
+    # Create sockets and attach to server for each listen address
+    my @socks;
+    for my $listen (@{$self->{_listen_addrs}}) {
+        my $sock = $self->_create_socket($listen, $use_reuseport);
+        push @socks, $sock;
+        $f->use_socket($sock);
+    }
+    $self->{sock} = $socks[0];   # backward compat: primary socket
+    $self->{_socks} = \@socks;   # all sockets
 
     $f->set_keepalive($_) for grep defined, delete $self->{keepalive};
+    $f->set_reverse_proxy($_) for grep defined, delete $self->{reverse_proxy};
+    $f->set_proxy_protocol($_) for grep defined, delete $self->{proxy_protocol};
     $f->read_timeout($_) for grep defined, delete $self->{read_timeout};
     $f->header_timeout($_) for grep defined, delete $self->{header_timeout};
     $f->max_connection_reqs($_) for grep defined, delete $self->{max_connection_reqs};
@@ -256,6 +266,47 @@ sub _prepare {
     }
     $f->max_connections($_) for grep defined, delete $self->{max_connections};
 
+    # Build tls hash from flat options (for Plack -o tls_cert_file=... -o tls_key_file=...)
+    if (!$self->{tls}) {
+        if (my $cert = delete $self->{tls_cert_file}) {
+            my $key = delete $self->{tls_key_file}
+                or croak "tls_cert_file requires tls_key_file";
+            $self->{tls} = { cert_file => $cert, key_file => $key };
+        } elsif (my $key = delete $self->{tls_key_file}) {
+            croak "tls_key_file requires tls_cert_file";
+        }
+    } else {
+        # tls hash takes precedence; discard flat options
+        delete $self->{tls_cert_file};
+        delete $self->{tls_key_file};
+    }
+
+    # TLS configuration: apply to all listeners
+    if (my $tls = delete $self->{tls}) {
+        croak "tls must be a hash reference" unless ref $tls eq 'HASH';
+        croak "tls requires cert_file" unless $tls->{cert_file};
+        croak "tls requires key_file" unless $tls->{key_file};
+
+        # H2 is off by default; only enable if h2 => 1 was passed
+        if (delete $self->{h2}) {
+            $tls->{h2} = 1;
+        }
+
+        if ($f->has_tls()) {
+            for my $i (0 .. $#socks) {
+                $f->set_tls(listener => $i, %$tls);
+            }
+            $self->{quiet} or warn "Feersum [$$]: TLS enabled on "
+                . scalar(@socks) . " listener(s)\n";
+        } else {
+            croak "tls option requires Feersum compiled with TLS support (need picotls submodule + OpenSSL; see Alien::OpenSSL)";
+        }
+    } else {
+        if (delete $self->{h2}) {
+            croak "h2 requires TLS (provide tls_cert_file and tls_key_file, or a tls hash)";
+        }
+    }
+
     $self->{endjinn} = $f;
     return;
 }
@@ -292,13 +343,12 @@ sub run {
     $self->_start_pre_fork if $self->{pre_fork};
     EV::run;
     $self->{quiet} or warn "Feersum [$$]: done\n";
-    $self->DESTROY();
+    $self->_cleanup();
     return;
 }
 
 sub _fork_another {
     my ($self, $slot) = @_;
-    weaken $self;
 
     my $pid = fork;
     croak "failed to fork: $!" unless defined $pid;
@@ -308,24 +358,30 @@ sub _fork_another {
         delete $self->{_kids};
         delete $self->{pre_fork};
 
-        # With SO_REUSEPORT, each child creates its own socket
+        # With SO_REUSEPORT, each child creates its own sockets
         # This eliminates accept() contention for better scaling
         if ($self->{_use_reuseport}) {
             $self->{endjinn}->unlisten();
-            if ($self->{sock}) {
-                close($self->{sock})
-                    or croak "close parent socket in child: $! (fd leak risk)";
+            for my $old_sock (@{$self->{_socks} || []}) {
+                close($old_sock)
+                    or do { warn "close parent socket in child: $!"; POSIX::_exit(1); };
             }
-            my $sock = $self->_create_socket($self->{_listen_addr}, 1);
-            $self->{sock} = $sock;
-            $self->{endjinn}->use_socket($sock);
+            my @new_socks;
+            for my $listen (@{$self->{_listen_addrs}}) {
+                my $sock = $self->_create_socket($listen, 1);
+                push @new_socks, $sock;
+                $self->{endjinn}->use_socket($sock);
+            }
+            $self->{sock} = $new_socks[0];
+            $self->{_socks} = \@new_socks;
         }
 
         eval { EV::run; }; ## no critic (RequireCheckingReturnValueOfEval)
         carp $@ if $@;
-        POSIX::exit($@ ? -1 : 0); ## no critic (ProhibitMagicNumbers)
+        POSIX::_exit($@ ? 1 : 0);  # _exit avoids running parent's END blocks
     }
 
+    weaken $self;  # prevent circular ref with watcher callback
     $self->{_n_kids}++;
     $self->{_kids}[$slot] = EV::child $pid, 0, sub {
         my $w = shift;
@@ -340,15 +396,22 @@ sub _fork_another {
         # Without SO_REUSEPORT, parent needs to accept during respawn
         unless ($self->{_use_reuseport}) {
             my $feersum = $self->{endjinn};
-            my $fd = fileno $self->{sock};
-            if (defined $fd) {
-                $feersum->accept_on_fd($fd);
+            my @socks = @{$self->{_socks} || [$self->{sock}]};
+            my $all_valid = 1;
+            for my $sock (@socks) {
+                unless (defined fileno $sock) {
+                    $all_valid = 0;
+                    last;
+                }
+            }
+            if ($all_valid) {
+                for my $sock (@socks) {
+                    $feersum->accept_on_fd(fileno $sock);
+                }
                 $self->_fork_another($slot);
                 $feersum->unlisten;
             } else {
-                # Socket invalid but still respawn to maintain pool size
-                carp "fileno returned undef during respawn, respawning anyway";
-                $self->_fork_another($slot);
+                carp "fileno returned undef during respawn, cannot respawn worker";
             }
         }
         else {
@@ -373,15 +436,15 @@ sub _start_pre_fork {
     # Parent stops accepting - children handle connections
     $self->{endjinn}->unlisten();
 
-    # With SO_REUSEPORT, parent can close its socket entirely
+    # With SO_REUSEPORT, parent can close its sockets entirely
     # Children have their own sockets
-    if ($self->{_use_reuseport} && $self->{sock}) {
-        if (close($self->{sock})) {
-            $self->{sock} = undef;
-        } else {
-            carp "close socket: $!";
-            # Keep reference so destructor can retry cleanup
+    if ($self->{_use_reuseport}) {
+        for my $sock (@{$self->{_socks} || []}) {
+            close($sock)
+                or do { warn "close parent socket after fork: $!"; POSIX::_exit(1); };
         }
+        $self->{sock} = undef;
+        $self->{_socks} = [];
     }
     return;
 }
@@ -402,10 +465,10 @@ sub quit {
     }
     else {
         # in child or solo process
-        $self->{endjinn}->graceful_shutdown(sub { POSIX::exit(0) });
+        $self->{endjinn}->graceful_shutdown(sub { POSIX::_exit(0) });
     }
 
-    $self->{_death} = EV::timer $death, 0, sub { POSIX::exit(1) };
+    $self->{_death} = EV::timer $death, 0, sub { POSIX::_exit(1) };
     return;
 }
 
@@ -444,10 +507,15 @@ invocation.
 
 =item listen
 
-Listen address as an arrayref containing a single address string, e.g.,
+Listen address as an arrayref containing one or more address strings, e.g.,
 C<< listen => ['localhost:5000'] >>. Formats: C<host:port> for IPv4,
-C<[host]:port> for IPv6 (e.g., C<['[::1]:8080']>). IPv6 requires C<reuseport>
-mode and Perl 5.14+. Alternatively, use C<host> and C<port> parameters.
+C<[host]:port> for IPv6 (e.g., C<['[::1]:8080']>).
+
+B<Important:> IPv6 addresses require both C<reuseport> mode to be enabled
+AND Perl 5.14+ with Socket IPv6 support. Without C<reuseport>, only IPv4
+addresses are supported.
+
+Alternatively, use C<host> and C<port> parameters.
 
 =item pre_fork
 
@@ -460,10 +528,67 @@ the C<run()> method).
 
 Enable/disable http keepalive requests.
 
+=item reverse_proxy
+
+Enable reverse proxy mode. When enabled, Feersum trusts X-Forwarded-For and
+X-Forwarded-Proto headers from upstream proxies:
+
+=over
+
+=item * REMOTE_ADDR is set to the first IP in X-Forwarded-For (the original client)
+
+=item * psgi.url_scheme is set from X-Forwarded-Proto (http or https)
+
+=back
+
+For the native interface, use C<< $req->client_address >> and C<< $req->url_scheme >>
+to get the forwarded values (these respect reverse_proxy mode automatically).
+
+Only enable this when Feersum is behind a trusted proxy that sets these headers.
+
+=item proxy_protocol
+
+Enable PROXY protocol support (HAProxy protocol). When enabled, Feersum expects
+each new connection to begin with a PROXY protocol header before any HTTP data.
+Both v1 (text) and v2 (binary) formats are auto-detected and supported.
+
+The PROXY protocol header provides the real client IP address when Feersum is
+behind a load balancer like HAProxy, AWS ELB/NLB, or nginx (with proxy_protocol).
+
+When a valid PROXY header is received:
+
+=over
+
+=item * REMOTE_ADDR/REMOTE_PORT are updated to the source address from the header
+
+=item * For v1 UNKNOWN or v2 LOCAL commands, the original address is preserved
+
+=back
+
+This option works independently of C<reverse_proxy>. When both are enabled,
+the PROXY protocol sets the base address, which can then be overridden by
+X-Forwarded-For headers if reverse_proxy is also enabled.
+
+B<Important:> Only enable this when ALL connections come through a proxy that
+sends PROXY headers. Connections without valid PROXY headers will be rejected
+with HTTP 400.
+
+Example HAProxy configuration:
+
+    backend feersum_backend
+        mode http
+        server feersum 127.0.0.1:5000 send-proxy-v2
+
 =item read_timeout
 
 Set read/keepalive timeout in seconds.
 
+=item header_timeout
+
+Set maximum time (in seconds) to receive complete HTTP headers (Slowloris
+protection). Default is 0 (disabled). When enabled, connections that don't
+complete headers within the timeout are closed.
+
 =item max_connection_reqs
 
 Set max requests per connection in case of keepalive - 0(default) for unlimited.
@@ -480,13 +605,65 @@ Set maximum concurrent connections (default: 0, unlimited). When the limit is
 reached, new connections are immediately closed. Provides protection against
 Slowloris-style DoS attacks.
 
+=item read_priority
+
+=item write_priority
+
+=item accept_priority
+
+Set libev I/O watcher priorities for read, write, and accept operations.
+Valid range is -2 (lowest) to +2 (highest), default is 0.
+
+=item tls
+
+Enable TLS 1.3 on all listeners. Pass a hash reference with C<cert_file>
+and C<key_file> paths:
+
+    Feersum::Runner->new(
+        listen => ['0.0.0.0:8443'],
+        tls    => { cert_file => 'server.crt', key_file => 'server.key' },
+        app    => $app,
+    )->run;
+
+Requires Feersum to be compiled with TLS support (picotls submodule + L<Alien::OpenSSL>).
+HTTP/2 is not enabled by default; pass C<< h2 => 1 >> separately to enable it.
+
+=item tls_cert_file
+
+=item tls_key_file
+
+Flat alternatives to the C<tls> hash, useful with Plack's C<-o> flag:
+
+    plackup -s Feersum -o tls_cert_file=server.crt -o tls_key_file=server.key
+
+Both must be specified together. If a C<tls> hash is also provided, it takes
+precedence and these are ignored.
+
+=item h2
+
+Enable HTTP/2 negotiation via ALPN on TLS listeners (default: off). Requires
+TLS to be enabled. When L<Alien::nghttp2> was available at build time and this
+option is set, HTTP/2 is negotiated alongside HTTP/1.1 during the TLS handshake.
+
+    Feersum::Runner->new(
+        listen => ['0.0.0.0:8443'],
+        tls    => { cert_file => 'server.crt', key_file => 'server.key' },
+        h2     => 1,
+        app    => $app,
+    )->run;
+
+If set without TLS enabled, a fatal error (croak) is raised.
+
 =item reuseport
 
 Enable SO_REUSEPORT for better prefork scaling (default: off). When enabled in
 combination with C<pre_fork>, each worker process creates its own socket bound
 to the same address. The kernel then distributes incoming connections across
 workers, eliminating accept() contention and improving multi-core scaling.
-Requires Linux 3.9+ or similar kernel support. Also enables IPv6 support.
+Requires Linux 3.9+ or similar kernel support.
+
+B<Note:> IPv6 support requires C<reuseport> to be enabled. Without reuseport,
+only IPv4 addresses are supported.
 
 =item epoll_exclusive
 
